#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


class ReadException(Exception):
    """
    Raise when there is an error reading data from an API Source
    """


class InvalidConnectorDefinitionException(Exception):
    """
    Raise when the connector definition is invalid
    """



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Iterable, Optional

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.requesters.request_options.request_options_provider import RequestOptionsProvider
from airbyte_cdk.sources.declarative.types import Record, StreamSlice, StreamState
from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class StreamSlicer(RequestOptionsProvider, JsonSchemaMixin):
    """
    Slices the stream into a subset of records.
    Slices enable state checkpointing and data retrieval parallelization.

    The stream slicer keeps track of the cursor state as a dict of cursor_field -> cursor_value

    See the stream slicing section of the docs for more information.
    """

    @abstractmethod
    def stream_slices(self, sync_mode: SyncMode, stream_state: StreamState) -> Iterable[StreamSlice]:
        """
        Defines stream slices

        :param sync_mode: The sync mode used the read data
        :param stream_state: The current stream state
        :return: List of stream slices
        """

    @abstractmethod
    def update_cursor(self, stream_slice: StreamSlice, last_record: Optional[Record] = None):
        """
        State setter, accept state serialized by state getter.

        :param stream_slice: Current stream_slice
        :param last_record: Last record read from the source
        """

    @abstractmethod
    def get_stream_state(self) -> StreamState:
        """Returns the current stream state"""



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

# Initialize Streams Package
from .exceptions import UserDefinedBackoffException
from .http import HttpStream, HttpSubStream

__all__ = ["HttpStream", "HttpSubStream", "UserDefinedBackoffException"]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from unittest.mock import MagicMock

import pytest as pytest
import requests
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.requesters.http_requester import HttpMethod, HttpRequester


def test_http_requester():
    http_method = "GET"

    request_options_provider = MagicMock()
    request_params = {"param": "value"}
    request_body_data = "body_key_1=value_1&body_key_2=value2"
    request_body_json = {"body_field": "body_value"}
    request_options_provider.get_request_params.return_value = request_params
    request_options_provider.get_request_body_data.return_value = request_body_data
    request_options_provider.get_request_body_json.return_value = request_body_json

    request_headers_provider = MagicMock()
    request_headers = {"header": "value"}
    request_headers_provider.get_request_headers.return_value = request_headers

    authenticator = MagicMock()

    error_handler = MagicMock()
    max_retries = 10
    backoff_time = 1000
    response_status = MagicMock()
    response_status.retry_in.return_value = 10
    error_handler.max_retries = max_retries
    error_handler.interpret_response.return_value = response_status
    error_handler.backoff_time.return_value = backoff_time

    config = {"url": "https://airbyte.io"}
    stream_slice = {"id": "1234"}

    name = "stream_name"

    requester = HttpRequester(
        name=name,
        url_base=InterpolatedString.create("{{ config['url'] }}", options={}),
        path=InterpolatedString.create("v1/{{ stream_slice['id'] }}", options={}),
        http_method=http_method,
        request_options_provider=request_options_provider,
        authenticator=authenticator,
        error_handler=error_handler,
        config=config,
        options={},
    )

    assert requester.get_url_base() == "https://airbyte.io/"
    assert requester.get_path(stream_state={}, stream_slice=stream_slice, next_page_token={}) == "v1/1234"
    assert requester.get_authenticator() == authenticator
    assert requester.get_method() == HttpMethod.GET
    assert requester.get_request_params(stream_state={}, stream_slice=None, next_page_token=None) == request_params
    assert requester.get_request_body_data(stream_state={}, stream_slice=None, next_page_token=None) == request_body_data
    assert requester.get_request_body_json(stream_state={}, stream_slice=None, next_page_token=None) == request_body_json
    assert requester.interpret_response_status(requests.Response()) == response_status
    assert {} == requester.request_kwargs(stream_state={}, stream_slice=None, next_page_token=None)


@pytest.mark.parametrize(
    "test_name, base_url, expected_base_url",
    [
        ("test_no_trailing_slash", "https://example.com", "https://example.com/"),
        ("test_with_trailing_slash", "https://example.com/", "https://example.com/"),
        ("test_with_v1_no_trailing_slash", "https://example.com/v1", "https://example.com/v1/"),
        ("test_with_v1_with_trailing_slash", "https://example.com/v1/", "https://example.com/v1/"),
    ],
)
def base_url_has_a_trailing_slash(test_name, base_url, expected_base_url):
    requester = HttpRequester(
        name="name",
        url_base=base_url,
        path="deals",
        http_method=HttpMethod.GET,
        request_options_provider=MagicMock(),
        authenticator=MagicMock(),
        error_handler=MagicMock(),
        config={},
        options={},
    )
    assert requester.get_url_base() == expected_base_url


@pytest.mark.parametrize(
    "test_name, base_url, expected_base_url",
    [
        ("test_no_leading_slash", "deals", "deals"),
        ("test_with_leading_slash", "/deals", "deals"),
        ("test_with_v1_no_leading_slash", "v1/deals", "v1/deals"),
        ("test_with_v1_with_trailing_slash", "/v1/deals", "v1/deals"),
    ],
)
def path_has_no_leading_slash(test_name, path, expected_path):
    requester = HttpRequester(
        name="name",
        url_base="https://example.com",
        path=path,
        http_method=HttpMethod.GET,
        request_options_provider=MagicMock(),
        authenticator=MagicMock(),
        error_handler=MagicMock(),
        config={},
        options={},
    )
    assert requester.get_path(stream_state={}, stream_slice={}, next_page_token={}) == expected_path



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.requesters.request_option import RequestOption, RequestOptionType


@pytest.mark.parametrize(
    "test_name, option_type, field_name, should_raise",
    [
        ("test_limit_path_no_field_name", RequestOptionType.path, None, False),
        ("test_limit_path_with_field_name", RequestOptionType.path, "field", True),
        ("test_limit_param_no_field_name", RequestOptionType.request_parameter, None, True),
        ("test_limit_param_with_field_name", RequestOptionType.request_parameter, "field", False),
        ("test_limit_header_no_field_name", RequestOptionType.header, None, True),
        ("test_limit_header_with_field_name", RequestOptionType.header, "field", False),
        ("test_limit_data_no_field_name", RequestOptionType.body_data, None, True),
        ("test_limit_data_with_field_name", RequestOptionType.body_data, "field", False),
        ("test_limit_json_no_field_name", RequestOptionType.body_json, None, True),
        ("test_limit_json_with_field_name", RequestOptionType.body_json, "field", False),
    ],
)
def test_request_option(test_name, option_type, field_name, should_raise):
    try:
        request_option = RequestOption(inject_into=option_type, field_name=field_name, options={})
        if should_raise:
            assert False
        assert request_option.field_name == field_name
        assert request_option.inject_into == option_type
    except ValueError:
        if not should_raise:
            assert False



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString

config = {"field": "value"}
options = {"hello": "world"}
kwargs = {"c": "airbyte"}


@pytest.mark.parametrize(
    "test_name, input_string, expected_value",
    [
        ("test_static_value", "HELLO WORLD", "HELLO WORLD"),
        ("test_eval_from_options", "{{ options['hello'] }}", "world"),
        ("test_eval_from_config", "{{ config['field'] }}", "value"),
        ("test_eval_from_kwargs", "{{ kwargs['c'] }}", "airbyte"),
        ("test_eval_from_kwargs", "{{ kwargs['c'] }}", "airbyte"),
    ],
)
def test_interpolated_string(test_name, input_string, expected_value):
    s = InterpolatedString.create(input_string, options=options)
    assert s.eval(config, **{"kwargs": kwargs}) == expected_value



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, List, Mapping, Union

import dpath.util
import requests
from airbyte_cdk.sources.declarative.decoders.decoder import Decoder
from airbyte_cdk.sources.declarative.decoders.json_decoder import JsonDecoder
from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.types import Config, Record
from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class DpathExtractor(RecordExtractor, JsonSchemaMixin):
    """
    Record extractor that searches a decoded response over a path defined as an array of fields.

    If the field pointer points to an array, that array is returned.
    If the field pointer points to an object, that object is returned wrapped as an array.
    If the field pointer points to an empty object, an empty array is returned.
    If the field pointer points to a non-existing path, an empty array is returned.

    Examples of instantiating this transform:
    ```
      extractor:
        type: DpathExtractor
        field_pointer:
          - "root"
          - "data"
    ```

    ```
      extractor:
        type: DpathExtractor
        field_pointer:
          - "root"
          - "{{ options['field'] }}"
    ```

    ```
      extractor:
        type: DpathExtractor
        field_pointer: []
    ```

    Attributes:
        transform (Union[InterpolatedString, str]): Pointer to the field that should be extracted
        config (Config): The user-provided configuration as specified by the source's spec
        decoder (Decoder): The decoder responsible to transfom the response in a Mapping
    """

    field_pointer: List[Union[InterpolatedString, str]]
    config: Config
    options: InitVar[Mapping[str, Any]]
    decoder: Decoder = JsonDecoder(options={})

    def __post_init__(self, options: Mapping[str, Any]):
        for pointer_index in range(len(self.field_pointer)):
            if isinstance(self.field_pointer[pointer_index], str):
                self.field_pointer[pointer_index] = InterpolatedString.create(self.field_pointer[pointer_index], options=options)

    def extract_records(self, response: requests.Response) -> List[Record]:
        response_body = self.decoder.decode(response)
        if len(self.field_pointer) == 0:
            extracted = response_body
        else:
            pointer = [pointer.eval(self.config) for pointer in self.field_pointer]
            extracted = dpath.util.get(response_body, pointer, default=[])
        if isinstance(extracted, list):
            return extracted
        elif extracted:
            return [extracted]
        else:
            return []



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.models import AirbyteCatalog, SyncMode


class CatalogHelper:
    @staticmethod
    def coerce_catalog_as_full_refresh(catalog: AirbyteCatalog) -> AirbyteCatalog:
        """
        Updates the sync mode on all streams in this catalog to be full refresh
        """
        coerced_catalog = catalog.copy()
        for stream in catalog.streams:
            stream.source_defined_cursor = False
            stream.supported_sync_modes = [SyncMode.full_refresh]
            stream.default_cursor_field = None

        # remove nulls
        return AirbyteCatalog.parse_raw(coerced_catalog.json(exclude_unset=True, exclude_none=True))



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from .oauth import Oauth2Authenticator, SingleUseRefreshTokenOauth2Authenticator
from .token import BasicHttpAuthenticator, MultipleTokenAuthenticator, TokenAuthenticator

__all__ = [
    "Oauth2Authenticator",
    "SingleUseRefreshTokenOauth2Authenticator",
    "TokenAuthenticator",
    "MultipleTokenAuthenticator",
    "BasicHttpAuthenticator",
]
